
#include "stdio.h"

#include "ad++.h"

#include "fft2d.h"


void read_image(char* filename, const Section2<int>& image,
                int* nx, int* ny, int* maxin) ;

void write_image(char* filename, const Section2<int>& image,
                 int nx, int ny, int maxin) ;

void fft_slice(const Section1<complex>& a, int isgn) ;


// _____________________________________________________________________
//|                                                                     |
//| Driver for 2 dimensional FFT program, in HPF.                       |
//|                                                                     |
//| H W Yau.  25th of January, 1994.                                    |
//| Northeast Parallel Architectures Center.                            |
//| Syracuse University.                                                |
//|_____________________________________________________________________|
//|                                                                     |
//| Edit record:                                                        |
//| 25/Jan/1995: First edit, code modified from Bhaven Avalani's.       |
//| 31/Jan/1995: Rewrote, to get rid of superfluous call to fft1d.      |
//| 02/Feb/1995: Included HPF constructs and directives.                |
//| 10/Feb/1995: Rewrote so fft_slice only sees a 1-D array.            |
//| 15/Feb/1995: Plugged in routines to FFT in a .pgm image.            |
//| 16/Feb/1995: Corrected Fourier mode removal code, now removes the   |
//|              central portion of the Fourier `image'.                |
//| 20/Feb/1995: Separated codes.  This is now FFT of a constant image. |
//| 17/Jun/1996: Added timer routine.                                   |
//| 10/Jul/1996: Added communications & computation times.              |
//| 20/Aug/1996: Added PURE attribute to fft_slice for PGI compiler to  |
//|              do INDEPENDENT parallelisation.                        |
//| 23/Sep/1996: Converted to C++/Adlib.  DBC.                          |
//| 27/Jan/1998: Rewritten for version 2.0 of Adlib.  DBC.              |
//|_____________________________________________________________________|


void main(int argc, char** argv) {
  AdlibInit(argc, argv) ;

  Procs1 p ;
  ON(p) {
    Range x(NN) ;
    BlockRange y(NN, p.dim(0)) ;

    Array2<int> image(x, y) ;
    Array2<complex> a(x, y) ;
    Array2<complex> tmp1(x, y) ;

    int sign, itrunc ;
    int nx, ny, maxin ;
    float norm ;

    float ti, tf ;
    float tstart, tend ;
    float tcomms, tcomp ;

    int n ;

    Index i(x), j(y) ;

// Read in from standard in, the number of Fourier modes to keep (in each
// dimension.

    gprintf("Please enter how many Fourier modes to remove"
	    " in each dimension.\n") ;
    gscanf("%d", &itrunc) ;

// Read in the image, and copy into array to be FFT-ed.

    read_image("wolf.pgm", image, &nx, &ny, &maxin) ;

    OVERALL(i) {
      OVERALL(j) {
	a(i, j) = complex(image(i, j)) ;
      } ALLOVER(j) ;
    } ALLOVER(i) ;

    norm = 1.0 / (nx * ny) ;

// Set parameter to tell fft_slice to do FFT transform.

    sign = 1 ;
 
    OVERALL(j)
      fft_slice(a.sect(x, j), sign) ;
    ALLOVER(j) ;

//______________________________________________________________________
//
//.. Transpose
//______________________________________________________________________

    transpose(tmp1, a) ;
    copy(a, tmp1) ;

//______________________________________________________________________
//

    OVERALL(j)
      fft_slice(a.sect(x, j), sign) ;
    ALLOVER(j) ;

//______________________________________________________________________
//
// Scale magnitude.

    OVERALL(i) {
      OVERALL(j) {
	a(i, j) *= norm ;
      } ALLOVER(j) ;
    } ALLOVER(i) ;

// Throw away modes in the middle of the frequency range.

    int ioff = (nx - itrunc) / 2 ;

    Index im(x.subrng(itrunc, ioff)) ;
    Index jm(y.subrng(itrunc, ioff)) ;
    OVERALL(im) {
      OVERALL(jm) {
	a(im, jm) = 0.0 ;
      } ALLOVER(jm) ;
    } ALLOVER(im) ;

//// Throw away Low frequency mode...
//
//    Index il(x.subrng(itrunc, 0)) ;
//    Index jl(y.subrng(itrunc, 0)) ;
//    OVERALL(il) {
//      OVERALL(jl) {
//	a(il, jl) = 0.0 ;
//      } ALLOVER(jl) ;
//    } ALLOVER(il) ;
//
//// Throw away High frequency modes
//
//    Index jh(y.subrng(itrunc, ny - itrunc + 1)) ;
//    OVERALL(i) {
//      OVERALL(jh) {
//	a(i, jh) = 0.0 ;
//      } ALLOVER(jh) ;
//    } ALLOVER(i) ;
//
//    Index ih(x.subrng(itrunc, nx - itrunc + 1)) ;
//    OVERALL(ih) {
//      OVERALL(j) {
//	a(ih, j) = 0.0 ;
//      } ALLOVER(j) ;
//    } ALLOVER(ih) ;


//_____________________________________________________________________
//
// Do Inverse transform.
//_____________________________________________________________________

    sign = -1 ;

    OVERALL(j)
      fft_slice(a.sect(x, j), sign) ;
    ALLOVER(j) ;

//______________________________________________________________________
//
//.. Transpose
//______________________________________________________________________

    transpose(tmp1, a) ;
    copy(a, tmp1) ;

//______________________________________________________________________
//

    OVERALL(j)
      fft_slice(a.sect(x, j), sign) ;
    ALLOVER(j) ;

//______________________________________________________________________
//
// Output the image.
//______________________________________________________________________


    OVERALL(i) {
      OVERALL(j) {
	image(i, j) = (int) a(i, j).abs() ;
      } ALLOVER(j) ;
    } ALLOVER(i) ;
    maxin = maxval(image) ;

    write_image("fft_out.pgm", image, nx, ny, maxin) ;

//    gprintf("Time taken is %12.3f seconds\n", tf) ;
//    gprintf("Number of Processors = %4i\n"
//            "Problem size = %5i\n"
//            "Communications = %9.3f\n"
//            "Compute        = %9.3f\n"
//            "Others         = %9.3f\n"
//            "Total time     = %9.3f\n",
//            nprocs, NN, tcomms, tcomp, (tf - tcomms - tcomp), tf) ;
  } NO(p) ;

  AdlibFinalize() ;
}

