
// Operations on block distributed ranges.


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `BlockRangeRep': Block-distributed template range.            //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class BlockRangeRep : public RangeRep {
public :
  BlockRangeRep(RangeRep* _kernel, const int _blockSize) :
          RangeRep(_blockSize * _kernel->size(),
                   _kernel, _kernel->lev() + 1, _kernel->dim()),
          blockSize(_blockSize) {
    kernel->incr() ;
  }

  ~BlockRangeRep() {
    kernel->decr() ;
  }

  RangeRep* subrng(const int _extent, const int _base, const int _stride) ;

  Format format() const ;

  int ker_glb(const int glb, int* sub) const ;

  int block(const int ker_glb, int* sub, int* count) const ;
  int block_step(int* sub_stp) const ;

  int width() const ;
  StrideRep* vector(StrideRep* _kernel, const int _length) const ;

  int glb(const int sub, const int ker_glb) const ;

  void show() ;
protected :
  int blockSize ;

MEMORY_POOL_DECS
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `BlockSubRangeRep': block-distributed subrange.               //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class BlockSubRangeRep : public RangeRep {
public :
  BlockSubRangeRep(const int _extent,
		   const int _base, const int _stride,
		   BlockRangeRep* _parent,
		   RangeRep* _kernel, const int _blockSize) :
          RangeRep(_extent, _kernel, _kernel->lev() + 1, _kernel->dim()),
          blockSize(_blockSize),
	  base(_base), stride(_stride), parent(_parent) {
    parent->incr() ;

    if(stride > 0) {
      ker_lo = base / blockSize ;
      ker_hi = (base + stride * (extent - 1)) / blockSize ;
    }
    else {
      ker_hi = base / blockSize ;
      ker_lo = (base + stride * (extent - 1)) / blockSize ;
    }
  }

  ~BlockSubRangeRep() {
    parent->decr() ;
  }

  RangeRep* subrng(const int _extent, const int _base, const int _stride) ;

  RangeRep* tem() ;
  int bas()  const ;
  int str()  const ;

  Format format() const ;

  int ker_glb(const int glb, int* sub) const ;

  int block(const int ker_glb, int* sub, int* count) const ;
  int block_step(int* sub_stp) const ;

  RangeRep* subker(int* ker_bas, int* ker_str) const ;

  int width() const ;
  StrideRep* vector(StrideRep* _kernel, const int _length) const ;

  int glb(const int sub, const int ker_glb) const ;

  void show() ;
protected :
  int blockSize ;
  int base, stride ;
  BlockRangeRep* parent ;

  int ker_lo, ker_hi ;

MEMORY_POOL_DECS
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// classes `PackedBlockRangeRep', `PackedBlockSubRangeRep':            //
//      Block-distributed ranges yielding arrays with ``packed''       //
//      layout.                                                        //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class PackedBlockRangeRep : public BlockRangeRep {
public :
  PackedBlockRangeRep(RangeRep* _kernel, const int _blockSize) :
          BlockRangeRep(_kernel, _blockSize) {}

  RangeRep* subrng(const int _extent, const int _base, const int _stride) ;
} ;


class PackedBlockSubRangeRep : public BlockSubRangeRep {
public :
  PackedBlockSubRangeRep(const int _extent,
                         const int _base, const int _stride,
                         BlockRangeRep* _parent,
                         RangeRep* _kernel, const int _blockSize) :
        BlockSubRangeRep(_extent, _base, _stride, _parent, _kernel, _blockSize),
        packing(_stride > 0 ? _stride : -_stride) {}

  RangeRep* subrng(const int _extent, const int _base, const int _stride) ;

  int width() const ;
  StrideRep* vector(StrideRep* _kernel, const int _length) const ;

protected :
  int packing ;

MEMORY_POOL_DECS
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `PackedStrideRep'.                                            //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class PackedStrideRep : public StrideRep {
public :
  PackedStrideRep(StrideRep* _kernel, const int _length, const int _packing) :
          StrideRep(_kernel, _length), packing(_packing) {}

  ~PackedStrideRep() {}

  int disp(const int sub) const ;
  int disp_step(const int sub_stp) const ;

protected :
  int packing ;

MEMORY_POOL_DECS
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// classes `ExtendedBlockRangeRep', `ExtendedBlockSubRangeRep':        //
//     Block-distributed range yielding arrays with ghost areas.       //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class ExtendedBlockRangeRep : public BlockRangeRep {
public :
  ExtendedBlockRangeRep(RangeRep* _kernel, const int _blockSize,
                        const int _wlo, const int _whi) :
          BlockRangeRep(_kernel, _blockSize), wlo(_wlo), whi(_whi) {}

  RangeRep* subrng(const int _extent, const int _base, const int _stride) ;

  int disp(const int sub) const ;

  int width() const ;
  StrideRep* vector(StrideRep* _kernel, const int _length) const ;

protected :
  int wlo, whi ;
  
MEMORY_POOL_DECS
} ;


class ExtendedBlockSubRangeRep : public BlockSubRangeRep {
public :
  ExtendedBlockSubRangeRep(const int _extent,
                           const int _base, const int _stride,
			   BlockRangeRep* _parent,
			   RangeRep* _kernel, const int _blockSize,
                           const int _wlo, const int _whi) :
    BlockSubRangeRep(_extent, _base, _stride, _parent, _kernel, _blockSize),
    wlo(_wlo), whi(_whi) {}

  RangeRep* subrng(const int _extent, const int _base, const int _stride) ;

  int width() const ;
  StrideRep* vector(StrideRep* _kernel, const int _length) const ;

protected :
  int wlo, whi ;
  
MEMORY_POOL_DECS
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `ExtendedStrideRep'.                                          //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class ExtendedStrideRep : public StrideRep {
public :
  ExtendedStrideRep(StrideRep* _kernel, const int _length, const int _wlo) :
          StrideRep(_kernel, _length), wlo(_wlo) {}

  ~ExtendedStrideRep() {}

  int disp(const int sub) const ;

protected :
  int wlo ;

MEMORY_POOL_DECS
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// classes `ADAPTBlockCyclicRangeRep', `ADAPTBlockCyclicSubRangeRep':  //
//     Block-cyclically distributed ranges yielding arrays with ADAPT  //
//     layout.                                                         //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class ADAPTBlockCyclicRangeRep : public PackedBlockRangeRep {
public :
  ADAPTBlockCyclicRangeRep(RangeRep* _kernel, const int _blockSize) :
          PackedBlockRangeRep(_kernel, _blockSize) {}

  RangeRep* subrng(const int _extent, const int _base, const int _stride) ;
} ;

class ADAPTBlockCyclicSubRangeRep : public PackedBlockSubRangeRep {
public :
  ADAPTBlockCyclicSubRangeRep(const int _extent,
                              const int _base, const int _stride,
                              BlockRangeRep* _parent,
                              RangeRep* _kernel, const int _blockSize) :
          PackedBlockSubRangeRep(_extent, _base, _stride,
                                 _parent, _kernel, _blockSize) {}

  RangeRep* subrng(const int _extent, const int _base, const int _stride) ;

  int volume() const ;
  StrideRep* stride_(int* size) const ;
} ;


// -------------------------------------------------------------
// Things to do
//
//   Revert to `BlockSubRangeRep' derived from `BlockRangeRep'.

