
// Operations on block distributed ranges.

#include <stdlib.h>
#include <stdio.h>
#include <memory.h>
#include "globals.h"
#include "Procs.h"
#include "Range.h"

inline int max(int x, int y) {return x > y ? x : y ; }
inline int min(int x, int y) {return x < y ? x : y ; }


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `BlockRangeRep': Block-distributed template range.            //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

RangeRep* BlockRangeRep :: subrng(const int _extent,
                                  const int _base, const int _stride) {
  return new BlockSubRangeRep(_extent, _base, _stride, this,
                              kernel, blockSize) ;
}

Format BlockRangeRep :: format() const {
  return DIST_BLOCK ;
}

int BlockRangeRep :: ker_glb(const int glb, int* sub) const {

  *sub = glb % blockSize ;
  return glb / blockSize ;
}

int BlockRangeRep :: block(const int ker_glb, int* sub, int* count) const {

  // `ker_glb' is a global subscript in the kernel range.

  *sub   = 0 ;
  *count = blockSize ;

  return ker_glb * blockSize ;
}

int BlockRangeRep :: block_step(int* sub_stp) const {

  *sub_stp = 1 ;
  return 1 ;
}

int BlockRangeRep :: width() const {return blockSize ; }

StrideRep* BlockRangeRep :: vector(StrideRep* _kernel,
                                   const int _length) const {
  return new StrideRep(_kernel, _length) ;
}

int BlockRangeRep :: glb(const int sub, const int ker_glb) const {
  return ker_glb * blockSize + sub ;
}

void BlockRangeRep :: show() {
  AdlibPrintf("BlockRangeRep :: show: size = %d ker = %p lev = %d dim = %p\n",
              extent, kernel, level, (DimensionRep*) dimension) ;
}

MEMORY_POOL_DEFN(BlockRangeRep)


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `BlockSubRangeRep': block-distributed subrange.               //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

RangeRep* BlockSubRangeRep :: subrng(const int _extent,
                                     const int _base, const int _stride) {
  return new BlockSubRangeRep(_extent,
			      base + stride * _base, stride * _stride,
			      parent, kernel, blockSize) ;
}

RangeRep* BlockSubRangeRep :: tem() {return parent ; }
int BlockSubRangeRep :: str() const {return stride ; }
int BlockSubRangeRep :: bas() const {return base ; }

Format BlockSubRangeRep :: format() const {
  return DIST_BLOCK ;
}

int BlockSubRangeRep :: ker_glb(const int glb, int* sub) const {
  const int tem = base + stride * glb ;

  *sub = tem % blockSize ;
  return tem / blockSize ;
}

int BlockSubRangeRep :: block(const int ker_glb, int* sub, int* count) const {

  if(ker_glb >= ker_lo && ker_glb <= ker_hi) {
    const int blockStart = ker_glb * blockSize ;

    const int blockStartOffset  = blockStart - base ;
    const int blockFinishOffset = blockStartOffset + blockSize - 1 ;

    int i_hi, i_lo ;
    if(stride > 0) {
      i_lo = max((blockStartOffset + stride - 1) / stride, 0) ;
      i_hi = min(blockFinishOffset / stride, extent - 1) ;
    }
    else {
      i_lo = max((blockFinishOffset + stride + 1) / stride, 0) ;
      i_hi = min(blockStartOffset / stride, extent - 1) ;
    }
   
    *sub   = base + stride * i_lo - blockStart ;
    *count = i_hi - i_lo + 1 ;

    return i_lo ;
  }
  else {
    *sub   = base ;
    *count = 0 ;

    return 0 ;
  }
}

int BlockSubRangeRep :: block_step(int* sub_stp) const {

  *sub_stp = stride ;
  return 1 ;
}

RangeRep* BlockSubRangeRep :: subker(int* ker_bas, int* ker_str) const {

  if(stride > 0) {
    *ker_bas = ker_lo ;
    *ker_str = 1 ;
  }
  else {
    *ker_bas = ker_hi ;
    *ker_str = -1 ;
  }

  return kernel->subrng(ker_hi - ker_lo + 1, *ker_bas, *ker_str) ;
}

int BlockSubRangeRep :: width() const {return blockSize ; }

StrideRep* BlockSubRangeRep :: vector(StrideRep* _kernel,
                                      const int _length) const {
  return new StrideRep(_kernel, _length) ;
}

int BlockSubRangeRep :: glb(const int sub, const int ker_glb) const {
  return (ker_glb * blockSize + sub - base) / stride ;
}

void BlockSubRangeRep :: show() {
  AdlibPrintf("BlockSubRangeRep :: show: size = %d ker = %p lev = %d dim = %p\n",
              extent, kernel, level, (DimensionRep*) dimension) ;
}

MEMORY_POOL_DEFN(BlockSubRangeRep)


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// classes `PackedBlockRangeRep', `PackedBlockSubRangeRep':            //
//      Block-distributed ranges yielding arrays with ``packed''       //
//      layout.                                                        //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

RangeRep* PackedBlockRangeRep :: subrng(const int _extent,
                                        const int _base, const int _stride) {
  return new PackedBlockSubRangeRep(_extent, _base, _stride,
                                    this, kernel, blockSize) ;
}

RangeRep* PackedBlockSubRangeRep :: subrng(const int _extent,
                                           const int _base, const int _stride) {
  return new PackedBlockSubRangeRep(_extent,
				    base + stride * _base, stride * _stride,
				    parent, kernel, blockSize) ;
}

StrideRep* PackedBlockSubRangeRep :: vector(StrideRep* _kernel,
                                            const int _length) const {
  return new PackedStrideRep(_kernel, _length, packing) ;
}

int PackedBlockSubRangeRep :: width() const {
  return (blockSize + packing - 1) / packing ;
}

MEMORY_POOL_DEFN(PackedBlockSubRangeRep)


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// classes `PackedStrideRep'                                           //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

int PackedStrideRep :: disp(const int sub) const {
  return length * (sub / packing) ;
}

int PackedStrideRep :: disp_step(const int sub_stp) const {
  return length * (sub_stp / packing) ;
}

MEMORY_POOL_DEFN(PackedStrideRep)


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// classes `ExtendedBlockRangeRep', `ExtendedBlockSubRangeRep':        //
//     Block-distributed range yielding arrays with ghost areas.       //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

RangeRep* ExtendedBlockRangeRep :: subrng(const int _extent,
                                          const int _base, const int _stride) {
  return new ExtendedBlockSubRangeRep(_extent, _base, _stride, this,
                                      kernel, blockSize, wlo, whi) ;
}

int ExtendedBlockRangeRep :: width() const {
  return wlo + blockSize + whi ;
}

StrideRep* ExtendedBlockRangeRep :: vector(StrideRep* _kernel,
                                           const int _length) const {
  return new ExtendedStrideRep(_kernel, _length, wlo) ;
}


RangeRep* ExtendedBlockSubRangeRep :: subrng(const int _extent,
                                             const int _base,
                                             const int _stride) {
  return new ExtendedBlockSubRangeRep(_extent,
				      base + stride * _base, stride * _stride,
				      parent, kernel, blockSize, wlo, whi) ;
}

int ExtendedBlockSubRangeRep :: width() const {
  return wlo + blockSize + whi ;
}

StrideRep* ExtendedBlockSubRangeRep :: vector(StrideRep* _kernel,
                                              const int _length) const {
  return new ExtendedStrideRep(_kernel, _length, wlo) ;
}


MEMORY_POOL_DEFN(ExtendedBlockRangeRep)

MEMORY_POOL_DEFN(ExtendedBlockSubRangeRep)


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// classes `ExtendedStrideRep'                                         //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

int ExtendedStrideRep :: disp(const int sub) const {
  return length * (sub + wlo) ;
}

MEMORY_POOL_DEFN(ExtendedStrideRep)


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// classes `ADAPTBlockCyclicRangeRep', `ADAPTBlockCyclicSubRangeRep':  //
//     Block-cyclically distributed ranges yielding arrays with ADAPT  //
//     layout.                                                         //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

RangeRep* ADAPTBlockCyclicRangeRep :: subrng(const int _extent,
                                        const int _base, const int _stride) {
  return new ADAPTBlockCyclicSubRangeRep(_extent, _base, _stride,
                                         this, kernel, blockSize) ;
}


RangeRep* ADAPTBlockCyclicSubRangeRep :: subrng(const int _extent,
					        const int _base,
					        const int _stride) {
  return new ADAPTBlockCyclicSubRangeRep(_extent,
					 base + stride * _base,
					 stride * _stride,
                                         parent, kernel, blockSize) ;
}

StrideRep* ADAPTBlockCyclicSubRangeRep :: stride_(int* size) const {
  const int topBlock = (stride > 0 ? base + stride * (extent - 1) : base) /
                       blockSize ;

  const int len1 = *size ;
  *size *= width() ;

  const int len0 = *size ;
  *size *= topBlock / dimension.size() + 1 ;

  return vector(kernel->vector(0, len0), len1) ;
}

int ADAPTBlockCyclicSubRangeRep :: volume() const {
  const int topBlock = (stride > 0 ? base + stride * (extent - 1) : base) /
                       blockSize ;

  return width() * (topBlock / dimension.size() + 1) ;
}


// -------------------------------------------------------------
// Things to do
//
//   `subker' returns non-empty subrange even for extent = 0
//        (not a problem?)

